#![feature(unboxed_closures)]
#![allow(unused_imports)]
#![allow(dead_code)]
#![allow(non_camel_case_types)]
#![feature(quote)]

extern crate collections;

use std::collections::lru_cache::LruCache;
use std::collections::hash_map::HashMap;
use collections::hash::Hash;
use std::mem;

pub trait MemoizeFn<Args, Result> : MemoizeFnBackend<Args, Result> where Args : Clone , Result : Clone {
    fn mem_call(&self, args: Args) -> Result {
        let mut _self : &mut Self = unsafe { mem::transmute(self) };
        
        match _self.get_evaluation(&args) {
            Some(result) => return result,
            None => {
                let result = _self.call_underlying_fn(args.clone());
                _self.store_evaluation(args, result.clone());
                result
            }
        }
    }

    fn call_underlying_fn(&self, args : Args) -> Result;
}

pub trait MemoizeFnBackend<Args, Result> {
    fn get_evaluation(&mut self, args : &Args) -> Option<Result>;
    fn store_evaluation(&mut self, args : Args, result : Result);
}

//Different Backends
//-----------------------------------------------------------------------------

//HashMap Backedend
pub struct HashMapMemoizer<Args, Result> where Args : Eq + Hash {
    evals : HashMap<Args, Result>
}

impl<Args, Result> HashMapMemoizer<Args, Result> where Args : Eq + Hash {
    pub fn new() -> HashMapMemoizer<Args, Result> { 
        HashMapMemoizer {
            evals : HashMap::new()
        }
    }
}

impl<Args, Result> MemoizeFnBackend<Args, Result> for HashMapMemoizer<Args, Result> where Args : Eq + Hash, Result : Clone {
    fn get_evaluation(&mut self, args : &Args) -> Option<Result>{
        match self.evals.get(args) {
            Some(result) => Some(result.clone()),
            None => None
        }
    }

    fn store_evaluation(&mut self, args : Args, result : Result) {
        self.evals.insert(args, result);
    }   
}

//LRU Cache Backend
pub struct LRUCacheMemoizer<Args, Result> where Args : Eq + Hash {
    evals : LruCache<Args, Result>
}

impl<Args, Result> LRUCacheMemoizer<Args, Result> where Args : Eq + Hash {
    pub fn new(size : uint) -> LRUCacheMemoizer<Args, Result> { 
        LRUCacheMemoizer {
            evals : LruCache::new(size)
        }
    }
}

impl<Args, Result> MemoizeFnBackend<Args, Result> for LRUCacheMemoizer<Args, Result> where Args : Eq + Hash, Result : Clone {
    fn get_evaluation(&mut self, args : &Args) -> Option<Result>{
        match self.evals.get(args) {
            Some(result) => Some(result.clone()),
            None => None
        }
    }

    fn store_evaluation(&mut self, args : Args, result : Result) {
        self.evals.insert(args, result);
    }   
}

//timed cache that lazily invalidates its contents based off their timestamp
//don't be proactive in purging them...get them and see if the timestamp is still valid


